@@ -7,8 +7,10 @@ from tempfile import TemporaryFile, NamedTemporaryFile
 import wave
 import sys
 import struct
-from .logging_utils import log_conversion
+from .logging_utils import log_conversion, log_subprocess_output
+from .utils import mediainfo_json, fsdecode
 import base64
+from collections import namedtuple
 
 try:
     from StringIO import StringIO
@@ -69,6 +71,7 @@ class ClassPropertyDescriptor(object):
         self.fset = func
         return self
 
+
 def classproperty(func):
     if not isinstance(func, (classmethod, staticmethod)):
         func = classmethod(func)
@@ -81,6 +84,66 @@ AUDIO_FILE_EXT_ALIASES = {
     "wave": "wav",
 }
 
+WavSubChunk = namedtuple('WavSubChunk', ['id', 'position', 'size'])
+WavData = namedtuple('WavData', ['audio_format', 'channels', 'sample_rate',
+                                 'bits_per_sample', 'raw_data'])
+
+
+def extract_wav_headers(data):
+    # def search_subchunk(data, subchunk_id):
+    pos = 12  # The size of the RIFF chunk descriptor
+    subchunks = []
+    while pos + 8 < len(data) and len(subchunks) < 10:
+        subchunk_id = data[pos:pos + 4]
+        subchunk_size = struct.unpack_from('<I', data[pos + 4:pos + 8])[0]
+        subchunks.append(WavSubChunk(subchunk_id, pos, subchunk_size))
+        if subchunk_id == b'data':
+            # 'data' is the last subchunk
+            break
+        pos += subchunk_size + 8
+
+    return subchunks
+
+
+def read_wav_audio(data, headers=None):
+    if not headers:
+        headers = extract_wav_headers(data)
+
+    fmt = [x for x in headers if x.id == b'fmt ']
+    if not fmt or fmt[0].size < 16:
+        raise CouldntDecodeError("Couldn't find fmt header in wav data")
+    fmt = fmt[0]
+    pos = fmt.position + 8
+    audio_format = struct.unpack_from('<H', data[pos:pos + 2])[0]
+    if audio_format != 1 and audio_format != 0xFFFE:
+        raise CouldntDecodeError("Unknown audio format 0x%X in wav data" %
+                                 audio_format)
+
+    channels = struct.unpack_from('<H', data[pos + 2:pos + 4])[0]
+    sample_rate = struct.unpack_from('<I', data[pos + 4:pos + 8])[0]
+    bits_per_sample = struct.unpack_from('<H', data[pos + 14:pos + 16])[0]
+
+    data_hdr = headers[-1]
+    if data_hdr.id != b'data':
+        raise CouldntDecodeError("Couldn't find data header in wav data")
+
+    pos = data_hdr.position + 8
+    return WavData(audio_format, channels, sample_rate, bits_per_sample,
+                   data[pos:pos + data_hdr.size])
+
+
+def fix_wav_headers(data):
+    headers = extract_wav_headers(data)
+    if not headers or headers[-1].id != b'data':
+        return
+
+    # Set the file size in the RIFF chunk descriptor
+    data[4:8] = struct.pack('<I', len(data) - 8)
+
+    # Set the data size in the data subchunk
+    pos = headers[-1].position
+    data[pos + 4:pos + 8] = struct.pack('<I', len(data) - pos - 8)
+
 
 class AudioSegment(object):
     """
@@ -146,25 +209,21 @@ class AudioSegment(object):
                 data = data if isinstance(data, (basestring, bytes)) else data.read()
             except(OSError):
                 d = b''
-                reader = data.read(2**31-1)
+                reader = data.read(2 ** 31 - 1)
                 while reader:
                     d += reader
-                    reader = data.read(2**31-1)
+                    reader = data.read(2 ** 31 - 1)
                 data = d
 
-            raw = wave.open(StringIO(data), 'rb')
+            wav_data = read_wav_audio(data)
+            if not wav_data:
+                raise CouldntDecodeError("Couldn't read wav audio from data")
 
-            raw.rewind()
-            self.channels = raw.getnchannels()
-            self.sample_width = raw.getsampwidth()
-            self.frame_rate = raw.getframerate()
+            self.channels = wav_data.channels
+            self.sample_width = wav_data.bits_per_sample // 8
+            self.frame_rate = wav_data.sample_rate
             self.frame_width = self.channels * self.sample_width
-
-            raw.rewind()
-
-            # the "or b''" base case is a work-around for a python 3.4
-            # see https://github.com/jiaaro/pydub/pull/107
-            self._data = raw.readframes(float('inf')) or b''
+            self._data = wav_data.raw_data
 
         # Convert 24-bit audio to 32-bit audio.
         # (stdlib audioop and array modules do not support 24-bit data)
@@ -185,7 +244,6 @@ class AudioSegment(object):
                 old_bytes = struct.pack(pack_fmt, b0, b1, b2)
                 byte_buffer.write(old_bytes)
 
-
             self._data = byte_buffer.getvalue()
             self.sample_width = 4
             self.frame_width = self.channels * self.sample_width
@@ -199,7 +257,6 @@ class AudioSegment(object):
         """
         return self._data
 
-
     def get_array_of_samples(self):
         """
         returns the raw_data as an array of samples
@@ -230,6 +287,12 @@ class AudioSegment(object):
 
     def __getitem__(self, millisecond):
         if isinstance(millisecond, slice):
+            if millisecond.step:
+                return (
+                    self[i:i + millisecond.step]
+                    for i in xrange(*millisecond.indices(len(self)))
+                )
+
             start = millisecond.start if millisecond.start is not None else 0
             end = millisecond.stop if millisecond.stop is not None \
                 else len(self)
@@ -404,7 +467,8 @@ class AudioSegment(object):
         segs = cls._sync(*mono_segments)
 
         if segs[0].channels != 1:
-            raise ValueError("AudioSegment.from_mono_audiosegments requires all arguments are mono AudioSegment instances")
+            raise ValueError(
+                "AudioSegment.from_mono_audiosegments requires all arguments are mono AudioSegment instances")
 
         channels = len(segs)
         sample_width = segs[0].sample_width
@@ -427,7 +491,7 @@ class AudioSegment(object):
         )
 
     @classmethod
-    def from_file(cls, file, format=None, codec=None, parameters=None, **kwargs):
+    def from_file_using_temporary_files(cls, file, format=None, codec=None, parameters=None, **kwargs):
         orig_file = file
         file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
 
@@ -441,11 +505,15 @@ class AudioSegment(object):
                 return True
             if isinstance(orig_file, basestring):
                 return orig_file.lower().endswith(".{0}".format(f))
+            if isinstance(orig_file, bytes):
+                return orig_file.lower().endswith((".{0}".format(f)).encode('utf8'))
             return False
 
         if is_format("wav"):
             try:
-                return cls._from_safe_wav(file)
+                obj = cls._from_safe_wav(file)
+                file.close()
+                return obj
             except:
                 file.seek(0)
         elif is_format("raw") or is_format("pcm"):
@@ -458,7 +526,9 @@ class AudioSegment(object):
                 'channels': channels,
                 'frame_width': channels * sample_width
             }
-            return cls(data=file.read(), metadata=metadata)
+            obj = cls(data=file.read(), metadata=metadata)
+            file.close()
+            return obj
 
         input_file = NamedTemporaryFile(mode='wb', delete=False)
         try:
@@ -466,13 +536,15 @@ class AudioSegment(object):
         except(OSError):
             input_file.flush()
             input_file.close()
-            input_file = NamedTemporaryFile(mode='wb', delete=False, buffering=2**31-1)
-            file = open(orig_file, buffering=2**13-1, mode='rb')
-            reader = file.read(2**31-1)
+            input_file = NamedTemporaryFile(mode='wb', delete=False, buffering=2 ** 31 - 1)
+            file.close()
+            file = open(orig_file, buffering=2 ** 13 - 1, mode='rb')
+            reader = file.read(2 ** 31 - 1)
             while reader:
                 input_file.write(reader)
-                reader = file.read(2**31-1)
+                reader = file.read(2 ** 31 - 1)
         input_file.flush()
+        file.close()
 
         output = NamedTemporaryFile(mode="rb", delete=False)
 
@@ -502,18 +574,131 @@ class AudioSegment(object):
 
         log_conversion(conversion_command)
 
-        p = subprocess.Popen(conversion_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        with open(os.devnull, 'rb') as devnull:
+            p = subprocess.Popen(conversion_command, stdin=devnull, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         p_out, p_err = p.communicate()
 
-        if p.returncode != 0:
-            raise CouldntDecodeError("Decoding failed. ffmpeg returned error code: {0}\n\nOutput from ffmpeg/avlib:\n\n{1}".format(p.returncode, p_err))
+        log_subprocess_output(p_out)
+        log_subprocess_output(p_err)
 
-        obj = cls._from_safe_wav(output)
+        try:
+            if p.returncode != 0:
+                raise CouldntDecodeError(
+                    "Decoding failed. ffmpeg returned error code: {0}\n\nOutput from ffmpeg/avlib:\n\n{1}".format(
+                        p.returncode, p_err))
+            obj = cls._from_safe_wav(output)
+        finally:
+            input_file.close()
+            output.close()
+            os.unlink(input_file.name)
+            os.unlink(output.name)
 
-        input_file.close()
-        output.close()
-        os.unlink(input_file.name)
-        os.unlink(output.name)
+        return obj
+
+    @classmethod
+    def from_file(cls, file, format=None, codec=None, parameters=None, **kwargs):
+        orig_file = file
+        try:
+            filename = fsdecode(file)
+        except TypeError:
+            filename = None
+        file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
+
+        if format:
+            format = format.lower()
+            format = AUDIO_FILE_EXT_ALIASES.get(format, format)
+
+        def is_format(f):
+            f = f.lower()
+            if format == f:
+                return True
+
+            if filename:
+                return filename.lower().endswith(".{0}".format(f))
+
+            return False
+
+        if is_format("wav"):
+            try:
+                return cls._from_safe_wav(file)
+            except:
+                file.seek(0)
+        elif is_format("raw") or is_format("pcm"):
+            sample_width = kwargs['sample_width']
+            frame_rate = kwargs['frame_rate']
+            channels = kwargs['channels']
+            metadata = {
+                'sample_width': sample_width,
+                'frame_rate': frame_rate,
+                'channels': channels,
+                'frame_width': channels * sample_width
+            }
+            return cls(data=file.read(), metadata=metadata)
+
+        conversion_command = [cls.converter,
+                              '-y',  # always overwrite existing files
+                              ]
+
+        # If format is not defined
+        # ffmpeg/avconv will detect it automatically
+        if format:
+            conversion_command += ["-f", format]
+
+        if codec:
+            # force audio decoder
+            conversion_command += ["-acodec", codec]
+
+        if filename:
+            conversion_command += ["-i", filename]
+            stdin_parameter = None
+            stdin_data = None
+        else:
+            conversion_command += ["-i", "-"]
+            stdin_parameter = subprocess.PIPE
+            stdin_data = file.read()
+
+        info = mediainfo_json(orig_file)
+        if info:
+            audio_streams = [x for x in info['streams']
+                             if x['codec_type'] == 'audio']
+            # This is a workaround for some ffprobe versions that always say
+            # that mp3/mp4/aac/webm/ogg files contain fltp samples
+            if (audio_streams[0].get('sample_fmt') == 'fltp' and
+                    (is_format("mp3") or is_format("mp4") or is_format("aac") or
+                     is_format("webm") or is_format("ogg"))):
+                bits_per_sample = 16
+            else:
+                bits_per_sample = audio_streams[0]['bits_per_sample']
+            acodec = 'pcm_s%dle' % bits_per_sample
+            conversion_command += ["-acodec", acodec]
+
+        conversion_command += [
+            "-vn",  # Drop any video streams if there are any
+            "-f", "wav",  # output options (filename last)
+            "-"
+        ]
+
+        if parameters is not None:
+            # extend arguments with arbitrary set
+            conversion_command.extend(parameters)
+
+        log_conversion(conversion_command)
+
+        p = subprocess.Popen(conversion_command, stdin=stdin_parameter,
+                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        p_out, p_err = p.communicate(input=stdin_data)
+
+        if p.returncode != 0 or len(p_out) == 0:
+            file.close()
+            raise CouldntDecodeError(
+                "Decoding failed. ffmpeg returned error code: {0}\n\nOutput from ffmpeg/avlib:\n\n{1}".format(
+                    p.returncode, p_err))
+
+        p_out = bytearray(p_out)
+        fix_wav_headers(p_out)
+        obj = cls._from_safe_wav(BytesIO(p_out))
+
+        file.close()
 
         return obj
 
@@ -535,20 +720,25 @@ class AudioSegment(object):
 
     @classmethod
     def from_raw(cls, file, **kwargs):
-        return cls.from_file(file, 'raw', sample_width=kwargs['sample_width'], frame_rate=kwargs['frame_rate'], channels=kwargs['channels'])
+        return cls.from_file(file, 'raw', sample_width=kwargs['sample_width'], frame_rate=kwargs['frame_rate'],
+                             channels=kwargs['channels'])
 
     @classmethod
     def _from_safe_wav(cls, file):
         file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
         file.seek(0)
-        return cls(data=file)
+        obj = cls(data=file)
+        file.close()
+        return obj
 
-    def export(self, out_f=None, format='mp3', codec=None, bitrate=None, parameters=None, tags=None, id3v2_version='4', cover=None):
+    def export(self, out_f=None, format='mp3', codec=None, bitrate=None, parameters=None, tags=None, id3v2_version='4',
+               cover=None):
         """
         Export an AudioSegment to a file with given options
 
         out_f (string):
-            Path to destination audio file
+            Path to destination audio file. Also accepts os.PathLike objects on
+            python >= 3.6
 
         format (string)
             Format for destination audio file.
@@ -620,9 +810,10 @@ class AudioSegment(object):
 
         if cover is not None:
             if cover.lower().endswith(('.png', '.jpg', '.jpeg', '.bmp', '.tif', '.tiff')) and format == "mp3":
-                conversion_command.extend(["-i" , cover, "-map", "0", "-map", "1", "-c:v", "mjpeg"])
+                conversion_command.extend(["-i", cover, "-map", "0", "-map", "1", "-c:v", "mjpeg"])
             else:
-                raise AttributeError("Currently cover images are only supported by MP3 files. The allowed image formats are: .tif, .jpg, .bmp, .jpeg and .png.")
+                raise AttributeError(
+                    "Currently cover images are only supported by MP3 files. The allowed image formats are: .tif, .jpg, .bmp, .jpeg and .png.")
 
         if codec is not None:
             # force audio encoder
@@ -651,7 +842,7 @@ class AudioSegment(object):
                         raise InvalidID3TagVersion(
                             "id3v2_version not allowed, allowed versions: %s" % id3v2_allowed_versions)
                     conversion_command.extend([
-                        "-id3v2_version",  id3v2_version
+                        "-id3v2_version", id3v2_version
                     ])
 
         if sys.platform == 'darwin':
@@ -664,11 +855,17 @@ class AudioSegment(object):
         log_conversion(conversion_command)
 
         # read stdin / write stdout
-        p = subprocess.Popen(conversion_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        with open(os.devnull, 'rb') as devnull:
+            p = subprocess.Popen(conversion_command, stdin=devnull, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         p_out, p_err = p.communicate()
 
+        log_subprocess_output(p_out)
+        log_subprocess_output(p_err)
+
         if p.returncode != 0:
-            raise CouldntEncodeError("Encoding failed. ffmpeg/avlib returned error code: {0}\n\nCommand:{1}\n\nOutput from ffmpeg/avlib:\n\n{2}".format(p.returncode, conversion_command, p_err))
+            raise CouldntEncodeError(
+                "Encoding failed. ffmpeg/avlib returned error code: {0}\n\nCommand:{1}\n\nOutput from ffmpeg/avlib:\n\n{2}".format(
+                    p.returncode, conversion_command, p_err))
 
         output.seek(0)
         out_f.write(output.read())
@@ -737,11 +934,13 @@ class AudioSegment(object):
         if channels == 2 and self.channels == 1:
             fn = audioop.tostereo
             frame_width = self.frame_width * 2
+            fac = 1
         elif channels == 1 and self.channels == 2:
             fn = audioop.tomono
             frame_width = self.frame_width // 2
+            fac = 0.5
 
-        converted = fn(self._data, self.sample_width, 1, 1)
+        converted = fn(self._data, self.sample_width, fac, fac)
 
         return self._spawn(data=converted,
                            overrides={
@@ -924,11 +1123,11 @@ class AudioSegment(object):
             if gain_during_overlay:
                 seg1_overlaid = seg1[pos:pos + seg2_len]
                 seg1_adjusted_gain = audioop.mul(seg1_overlaid, self.sample_width,
-                                               db_to_float(float(gain_during_overlay)))
+                                                 db_to_float(float(gain_during_overlay)))
                 output.write(audioop.add(seg1_adjusted_gain, seg2, sample_width))
             else:
                 output.write(audioop.add(seg1[pos:pos + seg2_len], seg2,
-                                     sample_width))
+                                         sample_width))
             pos += seg2_len
 
             # dec times to break our while loop (eventually)
@@ -962,7 +1161,9 @@ class AudioSegment(object):
         output.write(seg2[crossfade:]._data)
 
         output.seek(0)
-        return seg1._spawn(data=output)
+        obj = seg1._spawn(data=output)
+        output.close()
+        return obj
 
     def fade(self, to_gain=0, from_gain=0, start=None, end=None,
              duration=None):
@@ -1027,7 +1228,7 @@ class AudioSegment(object):
 
         # fades longer than 100ms can use coarse fading (one gain step per ms),
         # shorter fades will have audible clicks so they use precise fading
-        #(one gain step per sample)
+        # (one gain step per sample)
         if duration > 100:
             scale_step = gain_delta / duration
 
@@ -1074,14 +1275,15 @@ class AudioSegment(object):
         )
 
     def _repr_html_(self):
-            src = """
+        src = """
                     <audio controls>
                         <source src="data:audio/mpeg;base64,{base64}" type="audio/mpeg"/>
                         Your browser does not support the audio element.
                     </audio>
                   """
-            fh = self.export()
-            data = base64.b64encode(fh.read()).decode('ascii')
-            return src.format(base64=data)
+        fh = self.export()
+        data = base64.b64encode(fh.read()).decode('ascii')
+        return src.format(base64=data)
+
 
 from . import effects
